# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.recommenders.ensemble_recommender import WeightCache, EnsembleRecommender
from moto import mock_s3
import boto3
import json
from taar.recommenders.lazys3 import LazyJSONLoader
from .mocks import MockRecommenderFactory

EXPECTED = {'collaborative': 1000,
            'similarity': 100,
            'locale': 10}


def install_mock_ensemble_data(ctx):
    DATA = {'ensemble_weights': EXPECTED}

    S3_BUCKET = 'telemetry-parquet'
    ENSEMBLE_WEIGHTS = 'taar/ensemble/ensemble_weight.json'

    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=S3_BUCKET)
    conn.Object(S3_BUCKET, ENSEMBLE_WEIGHTS).put(Body=json.dumps(DATA))

    ctx['ensemble_weights'] = LazyJSONLoader(ctx,
                                             S3_BUCKET,
                                             ENSEMBLE_WEIGHTS)

    return ctx


@mock_s3
def test_weight_cache(test_ctx):
    ctx = install_mock_ensemble_data(test_ctx)
    wc = WeightCache(ctx)
    actual = wc.getWeights()
    assert EXPECTED == actual


@mock_s3
def test_recommendations(test_ctx):
    ctx = install_mock_ensemble_data(test_ctx)

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('def', 3320.0),
                        ('ijk', 3200.0),
                        ('hij', 3100.0),
                        ('lmn', 420.0)]

    factory = MockRecommenderFactory()
    ctx['recommender_factory'] = factory

    ctx['recommender_map'] = {'collaborative': factory.create('collaborative'),
                              'similarity': factory.create('similarity'),
                              'locale': factory.create('locale')}
    r = EnsembleRecommender(ctx.child())
    client = {'client_id': '12345'}  # Anything will work here

    recommendation_list = r.recommend(client, 5)
    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS


@mock_s3
def test_preinstalled_guids(test_ctx):
    ctx = install_mock_ensemble_data(test_ctx)

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('ijk', 3200.0),
                        ('lmn', 420.0),
                        ('klm', 409.99999999999994),
                        ('abc', 23.0)]

    factory = MockRecommenderFactory()
    ctx['recommender_factory'] = factory

    ctx['recommender_map'] = {'collaborative': factory.create('collaborative'),
                              'similarity': factory.create('similarity'),
                              'locale': factory.create('locale')}
    r = EnsembleRecommender(ctx.child())

    # 'hij' should be excluded from the suggestions list
    # The other two addon GUIDs 'def' and 'jkl' will never be
    # recommended anyway and should have no impact on results
    client = {'client_id': '12345',
              'installed_addons': ['def', 'hij', 'jkl']}

    recommendation_list = r.recommend(client, 5)
    print(recommendation_list)
    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import boto3
import json
from moto import mock_s3
from taar.recommenders import RecommendationManager
from taar.recommenders.recommendation_manager import TEST_CLIENT_IDS
from taar.recommenders.recommendation_manager import EMPTY_TEST_CLIENT_IDS
from taar.recommenders.lazys3 import LazyJSONLoader
from taar.schema import INTERVENTION_A
from taar.schema import INTERVENTION_B
from taar.schema import INTERVENTION_CONTROL
from taar.recommenders.base_recommender import AbstractRecommender
from .mocks import MockRecommenderFactory
from .test_hybrid_recommender import install_mock_curated_data


class StubRecommender(AbstractRecommender):
    """ A shared, stub recommender that can be used for testing.
    """
    def __init__(self, can_recommend, stub_recommendations):
        self._can_recommend = can_recommend
        self._recommendations = stub_recommendations

    def can_recommend(self, client_info, extra_data={}):
        return self._can_recommend

    def recommend(self, client_data, limit, extra_data={}):
        return self._recommendations


def install_mocks(ctx):
    ctx = ctx.child()

    class MockProfileFetcher:
        def get(self, client_id):
            return {'client_id': client_id}

    ctx['profile_fetcher'] = MockProfileFetcher()
    ctx['recommender_factory'] = MockRecommenderFactory()

    DATA = {'ensemble_weights': {'collaborative': 1000,
                                 'similarity': 100,
                                 'locale': 10}}

    S3_BUCKET = 'telemetry-parquet'
    ENSEMBLE_WEIGHTS = 'taar/ensemble/ensemble_weight.json'

    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=S3_BUCKET)
    conn.Object(S3_BUCKET, ENSEMBLE_WEIGHTS).put(Body=json.dumps(DATA))

    ctx['ensemble_weights'] = LazyJSONLoader(ctx,
                                             S3_BUCKET,
                                             ENSEMBLE_WEIGHTS)

    return ctx


@mock_s3
def test_none_profile_returns_empty_list(test_ctx):
    ctx = install_mocks(test_ctx)
    rec_manager = RecommendationManager(ctx)
    assert rec_manager.recommend("random-client-id", 10) == []


@mock_s3
def test_intervention_a(test_ctx):
    ctx = install_mocks(test_ctx)

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('def', 3320.0),
                        ('ijk', 3200.0),
                        ('hij', 3100.0),
                        ('lmn', 420.0),
                        ('klm', 409.99999999999994),
                        ('jkl', 400.0),
                        ('abc', 23.0),
                        ('fgh', 22.0),
                        ('efg', 21.0)]

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend('some_ignored_id',
                                            10,
                                            extra_data={'branch': INTERVENTION_A})

    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS


@mock_s3
def test_intervention_b(test_ctx):
    """The recommendation manager is currently very naive and just
    selects the first recommender which returns 'True' to
    can_recommend()."""

    ctx = install_mocks(test_ctx)
    ctx = install_mock_curated_data(ctx)

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend('some_ignored_id',
                                            4,
                                            extra_data={'branch': INTERVENTION_B})

    assert isinstance(recommendation_list, list)
    assert len(recommendation_list) == 4


@mock_s3
def test_intervention_control(test_ctx):
    ctx = install_mocks(test_ctx)
    ctx = install_mock_curated_data(ctx)

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend('some_ignored_id',
                                            10,
                                            extra_data={'branch': INTERVENTION_CONTROL})

    assert len(recommendation_list) == 0


@mock_s3
def test_fixed_client_id_valid(test_ctx):
    ctx = install_mocks(test_ctx)
    ctx = install_mock_curated_data(ctx)

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend(TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': INTERVENTION_A})

    assert len(recommendation_list) == 10

    recommendation_list = manager.recommend(TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': INTERVENTION_B})

    assert len(recommendation_list) == 10


@mock_s3
def test_intervention_names(test_ctx):
    # verify that intervention names can use hyphens instead of
    # underscore
    ctx = install_mocks(test_ctx)
    ctx = install_mock_curated_data(ctx)

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend(TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': 'intervention-a'})

    assert len(recommendation_list) == 10

    recommendation_list = manager.recommend(TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': 'intervention-b'})

    assert len(recommendation_list) == 10


@mock_s3
def test_fixed_client_id_empty_list(test_ctx):
    ctx = install_mocks(test_ctx)
    ctx = install_mock_curated_data(ctx)

    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend(EMPTY_TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': INTERVENTION_A})

    assert len(recommendation_list) == 0

    recommendation_list = manager.recommend(EMPTY_TEST_CLIENT_IDS[0],
                                            10,
                                            extra_data={'branch': INTERVENTION_B})

    assert len(recommendation_list) == 0



"""
These are global fixtures automagically loaded by pytest
"""

import pytest
from srgutil.context import default_context
from srgutil.interfaces import IClock

FAKE_LOCALE_DATA = {
    "te-ST": [
        "{1e6b8bce-7dc8-481c-9f19-123e41332b72}", "some-other@nice-addon.com",
        "{66d1eed2-a390-47cd-8215-016e9fa9cc55}", "{5f1594c3-0d4c-49dd-9182-4fbbb25131a7}"
    ],
    "en": [
        "some-uuid@test-addon.com", "other-addon@some-id.it"
    ]
}


@pytest.fixture
def test_ctx():
    ctx = default_context()
    ctx['clock'] = ctx[IClock]
    return ctx



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import colander


INTERVENTION_A = 'intervention_a'
INTERVENTION_B = 'intervention_b'
INTERVENTION_CONTROL = 'control'


class ExtraDataSchema(colander.MappingSchema):
    branch = colander.SchemaNode(colander.String(),
                                 validator=colander.OneOf([INTERVENTION_A,
                                                           INTERVENTION_B,
                                                           INTERVENTION_A.replace("_", '-'),
                                                           INTERVENTION_B.replace("_", '-'),
                                                           INTERVENTION_CONTROL]))


class RecommendationManagerQuerySchema(colander.MappingSchema):
    """
    This schema validates that arguments passed into
    `RecommendationManager::recommend` are of the correct type.

    Mostly useful for evoloving unittests and APIs in a stable way.
    """
    client_id = colander.SchemaNode(colander.String(),
                                    validator=colander.Length(max=200))
    limit = colander.SchemaNode(colander.Int(),
                                missing=10,
                                validator=colander.Range(min=1, max=100))
    extra_data = ExtraDataSchema()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test cases for the TAAR CollaborativeRecommender
"""

import numpy

from moto import mock_s3
import boto3
from taar.recommenders.collaborative_recommender import ITEM_MATRIX_CONFIG
from taar.recommenders.collaborative_recommender import ADDON_MAPPING_CONFIG
from taar.recommenders.collaborative_recommender import CollaborativeRecommender
from taar.recommenders.collaborative_recommender import positive_hash
from taar.recommenders.lazys3 import LazyJSONLoader
import json


"""
We need to generate a synthetic list of addons and relative weights
for co-occurance.  It's important to note that the
CollaborativeRecommender model expects that addon IDs are hashed using
the Java hash function.
"""


def install_none_mock_data(ctx):
    """
    Overload the 'real' addon model and mapping URLs responses so that
    we always get 404 errors.
    """
    conn = boto3.resource('s3', region_name='us-west-2')

    conn.create_bucket(Bucket=ITEM_MATRIX_CONFIG[0])
    conn.Object(ITEM_MATRIX_CONFIG[0], ITEM_MATRIX_CONFIG[1]).put(Body="")
    ctx['collaborative_item_matrix'] = LazyJSONLoader(ctx,
                                                      ITEM_MATRIX_CONFIG[0],
                                                      ITEM_MATRIX_CONFIG[1])

    # Don't reuse connections with moto.  badness happens
    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=ADDON_MAPPING_CONFIG[0])
    conn.Object(ADDON_MAPPING_CONFIG[0], ADDON_MAPPING_CONFIG[1]).put(Body="")
    ctx['collaborative_addon_mapping'] = LazyJSONLoader(ctx,
                                                        ADDON_MAPPING_CONFIG[0],
                                                        ADDON_MAPPING_CONFIG[1])
    return ctx


def install_mock_data(ctx):
    """
    Overload the 'real' addon model and mapping URLs responses so that
    we always the fixture data at the top of this test module.
    """

    addon_space = [{"id": "addon1.id", "name": "addon1.name", "isWebextension": True},
                   {"id": "addon2.id", "name": "addon2.name", "isWebextension": True},
                   {"id": "addon3.id", "name": "addon3.name", "isWebextension": True},
                   {"id": "addon4.id", "name": "addon4.name", "isWebextension": True},
                   {"id": "addon5.id", "name": "addon5.name", "isWebextension": True}]

    fake_addon_matrix = []
    for i, addon in enumerate(addon_space):
        row = {"id": positive_hash(addon['id']), "features": [0, 0.2, 0.0, 0.1, 0.15]}
        row['features'][i] = 1.0
        fake_addon_matrix.append(row)

    fake_mapping = {}
    for addon in addon_space:
        java_hash = positive_hash(addon['id'])
        fake_mapping[str(java_hash)] = addon

    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=ITEM_MATRIX_CONFIG[0])
    conn.Object(ITEM_MATRIX_CONFIG[0], ITEM_MATRIX_CONFIG[1]).put(Body=json.dumps(fake_addon_matrix))

    conn = boto3.resource('s3', region_name='us-west-2')
    conn.create_bucket(Bucket=ADDON_MAPPING_CONFIG[0])
    conn.Object(ADDON_MAPPING_CONFIG[0], ADDON_MAPPING_CONFIG[1]).put(Body=json.dumps(fake_mapping))

    ctx['collaborative_addon_mapping'] = LazyJSONLoader(ctx,
                                                        ADDON_MAPPING_CONFIG[0],
                                                        ADDON_MAPPING_CONFIG[1])
    ctx['collaborative_item_matrix'] = LazyJSONLoader(ctx,
                                                      ITEM_MATRIX_CONFIG[0],
                                                      ITEM_MATRIX_CONFIG[1])
    return ctx


@mock_s3
def test_cant_recommend(test_ctx):
    ctx = install_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})
    assert not r.can_recommend({"installed_addons": []})


@mock_s3
def test_can_recommend(test_ctx):
    ctx = install_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)

    # For some reason, moto doesn't like to play nice with this call
    # Check that we can recommend if we the user has at least an addon.
    assert r.can_recommend({"installed_addons": ["uBlock0@raymondhill.net"],
                            "client_id": "test-client"})


@mock_s3
def test_can_recommend_no_model(test_ctx):
    ctx = install_none_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)

    # We should never be able to recommend if something went wrong with the model.
    assert not r.can_recommend({})
    assert not r.can_recommend({"installed_addons": []})
    assert not r.can_recommend({"installed_addons": ["uBlock0@raymondhill.net"]})


@mock_s3
def test_empty_recommendations(test_ctx):
    # Tests that the empty recommender always recommends an empty list
    # of addons if we have no addons
    ctx = install_none_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)
    assert not r.can_recommend({})

    # Note that calling recommend() if can_recommend has failed is not
    # defined.


@mock_s3
def test_best_recommendation(test_ctx):
    # Make sure the structure of the recommendations is correct and that we
    # recommended the the right addon.
    ctx = install_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)

    # An non-empty set of addons should give a list of recommendations
    fixture_client_data = {"installed_addons": ["addon4.id"],
                           "client_id": "test_client"}
    assert r.can_recommend(fixture_client_data)
    recommendations = r.recommend(fixture_client_data, 1)

    assert isinstance(recommendations, list)
    assert len(recommendations) == 1

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[0]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon2.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.3225'))


@mock_s3
def test_recommendation_weights(test_ctx):
    """
    Weights should be ordered greatest to lowest
    """
    ctx = install_mock_data(test_ctx)
    r = CollaborativeRecommender(ctx)

    # An non-empty set of addons should give a list of recommendations
    fixture_client_data = {"installed_addons": ["addon4.id"],
                           "client_id": "test_client"}
    assert r.can_recommend(fixture_client_data)
    recommendations = r.recommend(fixture_client_data, 2)
    assert isinstance(recommendations, list)
    assert len(recommendations) == 2

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[0]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon2.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.3225'))

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[1]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon5.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.29'))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test cases for the TAAR Hybrid recommender
"""

from taar.recommenders.hybrid_recommender import CuratedRecommender
from taar.recommenders.hybrid_recommender import HybridRecommender
from taar.recommenders.ensemble_recommender import EnsembleRecommender

from taar.recommenders.hybrid_recommender import S3_BUCKET
from taar.recommenders.hybrid_recommender import CURATED_WHITELIST
# from taar.recommenders.hybrid_recommender import ENSEMBLE_WEIGHTS
from taar.recommenders.lazys3 import LazyJSONLoader
from .test_ensemblerecommender import install_mock_ensemble_data
from .mocks import MockRecommenderFactory

import json
from moto import mock_s3
import boto3


def install_no_curated_data(ctx):
    ctx = ctx.child()
    conn = boto3.resource('s3', region_name='us-west-2')

    conn.create_bucket(Bucket=S3_BUCKET)
    conn.Object(S3_BUCKET, CURATED_WHITELIST).put(Body="")
    ctx['curated_whitelist_data'] = LazyJSONLoader(ctx,
                                                   S3_BUCKET,
                                                   CURATED_WHITELIST)

    return ctx


def install_mock_curated_data(ctx):
    mock_data = []
    for i in range(20):
        mock_data.append({'GUID': str(i) * 16,
                          'Extension': 'WebExt %d' % i,
                          'Copy (final)': 'Copy for %d' % i})

    ctx = ctx.child()
    conn = boto3.resource('s3', region_name='us-west-2')

    conn.create_bucket(Bucket=S3_BUCKET)
    conn.Object(S3_BUCKET, CURATED_WHITELIST).put(Body=json.dumps(mock_data))
    ctx['curated_whitelist_data'] = LazyJSONLoader(ctx,
                                                   S3_BUCKET,
                                                   CURATED_WHITELIST)

    return ctx


def install_ensemble_fixtures(ctx):
    ctx = install_mock_ensemble_data(ctx)

    factory = MockRecommenderFactory()
    ctx['recommender_factory'] = factory

    ctx['recommender_map'] = {'collaborative': factory.create('collaborative'),
                              'similarity': factory.create('similarity'),
                              'locale': factory.create('locale')}
    ctx['ensemble_recommender'] = EnsembleRecommender(ctx.child())
    return ctx


@mock_s3
def test_curated_can_recommend(test_ctx):
    ctx = install_no_curated_data(test_ctx)
    r = CuratedRecommender(ctx)

    # CuratedRecommender will always recommend something no matter
    # what
    assert r.can_recommend({})
    assert r.can_recommend({"installed_addons": []})


@mock_s3
def test_curated_recommendations(test_ctx):
    ctx = install_mock_curated_data(test_ctx)
    r = CuratedRecommender(ctx)

    # CuratedRecommender will always recommend something no matter
    # what

    for LIMIT in range(1, 5):
        guid_list = r.recommend({'client_id': '000000'}, limit = LIMIT)
        # The curated recommendations should always return with some kind
        # of recommendations
        assert len(guid_list) == LIMIT


@mock_s3
def test_hybrid_recommendations(test_ctx):
    # verify that the recommendations mix the curated and
    # ensemble results
    ctx = install_mock_curated_data(test_ctx)
    ctx = install_ensemble_fixtures(ctx)

    r = HybridRecommender(ctx)

    # Test that we can generate lists of results
    for LIMIT in range(4, 8):
        guid_list = r.recommend({'client_id': '000000'}, limit = LIMIT)
        # The curated recommendations should always return with some kind
        # of recommendations
        assert len(guid_list) == LIMIT

    # Test that the results are actually mixed
    guid_list = r.recommend({'client_id': '000000'}, limit = 4)

    # A mixed list will have two recommendations with weight > 1.0
    # (ensemble) and 2 with exactly weight 1.0 from the curated list

    assert guid_list[0][1] > 1.0
    assert guid_list[1][1] > 1.0
    assert guid_list[2][1] == 1.0
    assert guid_list[3][1] == 1.0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from decouple import config
from srgutil.interfaces import IMozLogging
import boto3
import json
import zlib


DYNAMO_REGION = config('DYNAMO_REGION', default='us-west-2')
DYNAMO_TABLE_NAME = config('DYNAMO_TABLE_NAME', default='taar_addon_data_20180206')


class ProfileController:
    """
    This class provides basic read/write access into a AWS DynamoDB
    backed datastore.  The profile controller and profile fetcher code
    should eventually be merged as individually they don't "pull their
    weight".
    """

    def __init__(self, ctx, region_name, table_name):
        """
        Configure access to the DynamoDB instance
        """
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger('taar')
        self._ddb = boto3.resource('dynamodb', region_name=region_name)
        self._table = self._ddb.Table(table_name)

    def get_client_profile(self, client_id):
        """This fetches a single client record out of DynamoDB
        """
        try:
            response = self._table.get_item(Key={'client_id': client_id})
            compressed_bytes = response['Item']['json_payload'].value
            json_byte_data = zlib.decompress(compressed_bytes)
            json_str_data = json_byte_data.decode('utf8')
            return json.loads(json_str_data)
        except Exception as e:
            # Return None on error.  The caller in ProfileFetcher will
            # handle error logging
            msg = "Error loading client data for {}.  Error: {}"
            self.logger.error(msg.format(client_id, str(e)))
            return None


class ProfileFetcher:
    """ Fetch the latest information for a client on the backing
    datastore
    """
    def __init__(self, ctx):
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger('taar')
        self._client = ProfileController(self._ctx,
                                         region_name=DYNAMO_REGION,
                                         table_name=DYNAMO_TABLE_NAME)

    def set_client(self, client):
        self._client = client

    def get(self, client_id):
        profile_data = self._client.get_client_profile(client_id)

        if profile_data is None:
            self.logger.error("Client profile not found", extra={"client_id": client_id})
            return None

        addon_ids = [addon['addon_id']
                     for addon in profile_data.get('active_addons', [])
                     if not addon.get('is_system', False)]

        return {
            "client_id": client_id,
            "geo_city": profile_data.get("city", ''),
            "subsession_length": profile_data.get("subsession_length", 0),
            "locale": profile_data.get('locale', ''),
            "os": profile_data.get("os", ''),
            "installed_addons": addon_ids,
            "disabled_addons_ids": profile_data.get("disabled_addons_ids", []),
            "bookmark_count": profile_data.get("places_bookmarks_count", 0),
            "tab_open_count": profile_data.get("scalar_parent_browser_engagement_tab_open_event_count", 0),
            "total_uri": profile_data.get("scalar_parent_browser_engagement_total_uri_count", 0),
            "unique_tlds": profile_data.get("scalar_parent_browser_engagement_unique_domains_count", 0),
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .base_recommender import AbstractRecommender
from .lazys3 import LazyJSONLoader
from srgutil.interfaces import IMozLogging
import random
import operator as op

S3_BUCKET = 'telemetry-parquet'

ENSEMBLE_WEIGHTS = 'taar/ensemble/ensemble_weight.json'
CURATED_WHITELIST = 'telemetry-ml/addon_recommender/top_200_whitelist.json'


class CuratedWhitelistCache:
    """
    This fetches the curated whitelist from S3.

    A sample of the whitelist below :

        [{'GUID': guid_string,
          'Extension': extension_name,
          'Copy (final)': english_description},
        ]
    """
    def __init__(self, ctx):
        self._ctx = ctx
        if 'curated_whitelist_data' in self._ctx:
            self._data = self._ctx['curated_whitelist_data']
        else:
            self._data = LazyJSONLoader(self._ctx, S3_BUCKET, CURATED_WHITELIST)

    def get_whitelist(self):
        return self._data.get()[0]

    def get_randomized_guid_sample(self, item_count):
        """ Fetch a subset of randomzied GUIDs from the whitelist """
        dataset = self.get_whitelist()
        random.shuffle(dataset)
        samples = dataset[:item_count]
        return [s['GUID'] for s in samples]


class CuratedRecommender(AbstractRecommender):
    """
    The curated recommender just delegates to the whitelist
    that is provided by the AMO team.

    This recommender simply provides a randomized sample of
    pre-approved addons for recommendation. It does not use any other
    external data to generate recommendations, nor does it use any
    information from the Firefox agent.
    """

    def __init__(self, ctx):
        self._ctx = ctx

        self.logger = self._ctx[IMozLogging].get_logger('taar.curated')
        self._curated_wl = CuratedWhitelistCache(self._ctx)

    def can_recommend(self, client_data, extra_data={}):
        """The Curated recommender will always be able to recommend
        something"""
        self.logger.info("Curated can_recommend: {}".format(True))
        return True

    def recommend(self, client_data, limit, extra_data={}):
        """
        Curated recommendations are just random selections
        """
        guids = self._curated_wl.get_randomized_guid_sample(limit)

        results = [(guid, 1.0) for guid in guids]

        log_data = (client_data['client_id'], str(guids))
        self.logger.info("Curated recommendations client_id: [%s], guids: [%s]" % log_data)
        return results


class HybridRecommender(AbstractRecommender):
    """
    The EnsembleRecommender is a collection of recommenders where the
    results from each recommendation is amplified or dampened by a
    factor.  The aggregate results are combines and used to recommend
    addons for users.
    """
    def __init__(self, ctx):
        self._ctx = ctx

        self.logger = self._ctx[IMozLogging].get_logger('taar')

        self._ensemble_recommender = self._ctx['ensemble_recommender']
        self._curated_recommender = CuratedRecommender(self._ctx.child())

    def can_recommend(self, client_data, extra_data={}):
        """The ensemble recommender is always going to be
        available if at least one recommender is available"""
        ensemble_recommend = self._ensemble_recommender.can_recommend(client_data, extra_data)
        curated_recommend = self._curated_recommender.can_recommend(client_data, extra_data)
        result = ensemble_recommend and curated_recommend
        self.logger.info("Hybrid can_recommend: {}".format(result))
        return result

    def recommend(self, client_data, limit, extra_data={}):
        """
        Hybrid recommendations simply select half recommendations from
        the ensemble recommender, and half from the curated one.

        Duplicate recommendations are accomodated by rank ordering
        by weight.
        """

        preinstalled_addon_ids = client_data.get('installed_addons', [])

        # Compute an extended limit by adding the length of
        # the list of any preinstalled addons.
        extended_limit = limit + len(preinstalled_addon_ids)

        ensemble_suggestions = self._ensemble_recommender.recommend(client_data,
                                                                    extended_limit,
                                                                    extra_data)
        curated_suggestions = self._curated_recommender.recommend(client_data,
                                                                  extended_limit,
                                                                  extra_data)

        # Generate a set of results from each of the composite
        # recommenders.  We select one item from each recommender
        # sequentially so that we do not bias one recommender over the
        # other.
        merged_results = set()

        while len(merged_results) < limit and len(ensemble_suggestions) > 0 and len(curated_suggestions) > 0:

            r1 = ensemble_suggestions.pop()
            if r1[0] not in [temp[0] for temp in merged_results]:
                merged_results.add(r1)

            # Terminate early if we have an odd number for the limit
            if not (len(merged_results) < limit and
                    len(ensemble_suggestions) > 0 and
                    len(curated_suggestions) > 0):
                break

            r2 = curated_suggestions.pop()
            if r2[0] not in [temp[0] for temp in merged_results]:
                merged_results.add(r2)

        if len(merged_results) < limit:
            msg = "Defaulting to empty results. Insufficient recommendations found for client: %s" % client_data['client_id']
            self.logger.info(msg)
            return []

        sorted_results = sorted(list(merged_results),
                                key=op.itemgetter(1),
                                reverse=True)

        log_data = (client_data['client_id'],
                    str([r[0] for r in sorted_results]))

        self.logger.info("Hybrid recommendations client_id: [%s], guids: [%s]" % log_data)
        return sorted_results



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.recommenders.ensemble_recommender import EnsembleRecommender
from taar.recommenders.hybrid_recommender import HybridRecommender
from taar.schema import RecommendationManagerQuerySchema
from srgutil.interfaces import IMozLogging

import colander

from funcsigs import signature as inspect_sig
import funcsigs

from taar.schema import INTERVENTION_A
from taar.schema import INTERVENTION_B
from taar.schema import INTERVENTION_CONTROL

from taar.context import default_context

from .lazys3 import LazyJSONLoader
import random

# We need to build a default logger for the schema validation as there
# is no class to bind to yet.
ctx = default_context()
schema_logger = ctx[IMozLogging].get_logger('taar.schema_validate')


TEST_CLIENT_IDS = ['00000000-0000-0000-0000-000000000000',
                   '11111111-1111-1111-1111-111111111111',
                   '22222222-2222-2222-2222-222222222222',
                   '33333333-3333-3333-3333-333333333333']

EMPTY_TEST_CLIENT_IDS = ['00000000-aaaa-0000-0000-000000000000',
                         '11111111-aaaa-1111-1111-111111111111',
                         '22222222-aaaa-2222-2222-222222222222',
                         '33333333-aaaa-3333-3333-333333333333']


def schema_validate(colandar_schema):
    """
    Compute the function signature and apply a schema validator on the
    function.
    """
    def real_decorator(func):
        func_sig = inspect_sig(func)

        json_args = {}
        json_arg_names = []
        for key in func_sig.parameters.keys():
            json_arg_names.append(key)
            if key == 'self':
                continue

            default_val = func_sig.parameters[key].default
            if default_val is funcsigs._empty:
                json_args[key] = None
            else:
                json_args[key] = default_val

        def wrapper(*w_args, **w_kwargs):

            if json_arg_names[0] == 'self':
                # first arg is 'self', so this is a method.
                # Strip out self when doing argument validation
                for i, argval in enumerate(w_args[1:]):
                    kname = json_arg_names[i + 1]
                    json_args[kname] = argval
            else:
                for i, argval in enumerate(w_args):
                    kname = json_arg_names[i]
                    json_args[kname] = argval

            # Clobber the kwargs into the JSON version of the argument
            # list
            for k, v in w_kwargs.items():
                json_args[k] = v

            schema = RecommendationManagerQuerySchema()
            try:
                schema.deserialize(json_args)
            except colander.Invalid as e:
                msg = "Defaulting to empty results. Error deserializing input arguments: " + str(e.asdict().values())

                # This logger can't use the context logger as the code
                # is running in a method decorator
                schema_logger.warn(msg)
                # Invalid data means TAAR safely returns an empty list
                return []
            return func(*w_args, **w_kwargs)

        return wrapper
    return real_decorator


class RecommenderFactory:
    """
    A RecommenderFactory provides support to create recommenders.

    The existence of a factory enables injection of dependencies into
    the RecommendationManager and eases the implementation of test
    harnesses.
    """
    def __init__(self, ctx):
        self._ctx = ctx
        # This map is set in the default context
        self._recommender_factory_map = self._ctx['recommender_factory_map']

    def get_names(self):
        return self._recommender_factory_map.keys()

    def create(self, recommender_name):
        return self._recommender_factory_map[recommender_name]()


class RecommendationManager:
    """This class determines which of the set of recommendation
    engines will actually be used to generate recommendations."""

    def __init__(self, ctx):
        """Initialize the user profile fetcher and the recommenders.
        """
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger('taar')

        assert 'profile_fetcher' in self._ctx

        self.profile_fetcher = ctx['profile_fetcher']
        self._recommender_map = {}

        self.logger.info("Initializing recommenders")
        self._recommender_map[INTERVENTION_A] = EnsembleRecommender(self._ctx.child())

        hybrid_ctx = self._ctx.child()
        hybrid_ctx['ensemble_recommender'] = self._recommender_map[INTERVENTION_A]
        self._recommender_map[INTERVENTION_B] = HybridRecommender(hybrid_ctx)

        # The whitelist data is only used for test client IDs
        WHITELIST_S3_BUCKET = 'telemetry-parquet'
        WHITELIST_S3_KEY = 'telemetry-ml/addon_recommender/top_200_whitelist.json'
        self._whitelist_data = LazyJSONLoader(self._ctx, WHITELIST_S3_BUCKET, WHITELIST_S3_KEY)

    @schema_validate(RecommendationManagerQuerySchema)
    def recommend(self, client_id, limit, extra_data={}):
        """Return recommendations for the given client.

        The recommendation logic will go through each recommender and
        pick the first one that "can_recommend".

        :param client_id: the client unique id.
        :param limit: the maximum number of recommendations to return.
        :param extra_data: a dictionary with extra client data.
        """

        # Select recommendation output based on extra_data['branch']
        branch_selector = extra_data.get('branch', INTERVENTION_CONTROL)
        method_selector = branch_selector.replace('-', '_')
        method_name = 'recommend_{}'.format(method_selector)
        self.logger.info("Dispatching to method [{}]".format(method_name))
        branch_method = getattr(self, 'recommend_%s' % method_selector)

        if client_id in TEST_CLIENT_IDS:
            data = self._whitelist_data.get()[0]
            random.shuffle(data)
            samples = data[:limit]
            self.logger.info("Test ID detected [{}]".format(client_id))
            return [(s['GUID'], 1.1) for s in samples]

        if client_id in EMPTY_TEST_CLIENT_IDS:
            self.logger.info("Empty Test ID detected [{}]".format(client_id))
            return []

        client_info = self.profile_fetcher.get(client_id)
        if client_info is None:
            self.logger.warn("Defaulting to empty results.  No client info fetched from dynamo.")
            return []

        return branch_method(client_info, client_id, limit, extra_data)

    def recommend_intervention_a(self, client_info, client_id, limit, extra_data):
        """ Intervention A is the ensemble method """
        self.logger.info("Intervention A recommendation method invoked")
        recommender = self._recommender_map[INTERVENTION_A]
        return recommender.recommend(client_info, limit, extra_data)

    def recommend_intervention_b(self, client_info, client_id, limit, extra_data):
        """ Intervention A is the ensemble method hybridized with a
        curated list of addons """
        self.logger.info("Intervention B recommendation method invoked")
        recommender = self._recommender_map[INTERVENTION_B]
        return recommender.recommend(client_info, limit, extra_data)

    def recommend_control(self, client_info, client_id, limit, extra_data):
        """Run the control recommender - that is nothing"""
        self.logger.info("Control recommendation method invoked")
        return []



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar import ProfileFetcher
from taar.profile_fetcher import ProfileController
import boto3
import copy
import json
import zlib


class MockProfileController:
    def __init__(self, mock_profile):
        self._profile = mock_profile

    def get_client_profile(self, client_id):
        return self._profile


def test_profile_fetcher_returns_none(test_ctx):
    fetcher = ProfileFetcher(test_ctx)
    fetcher.set_client(MockProfileController(None))
    assert fetcher.get("random-client-id") is None


MOCK_DATA = {'profile': {u'scalar_parent_browser_engagement_total_uri_count': 791,
                         u'city': u'Rome',
                         u'scalar_parent_browser_engagement_tab_open_event_count': 46,
                         u'subsession_start_date': u'2017-09-20T10:00:00.0+02:00',
                         u'subsession_length': 3785,
                         u'places_bookmarks_count': 0,
                         u'scalar_parent_browser_engagement_unique_domains_count': 11,
                         u'os': u'Windows_NT',
                         u'active_addons': [{u'addon_id': u'e10srollout@mozilla.org'},
                                            {u'addon_id': u'firefox@getpocket.com'},
                                            {u'addon_id': u'webcompat@mozilla.org', 'is_system': True}],
                         u'locale': 'it-IT'},
             'expected_result': {"client_id": 'random-client-id',
                                 "bookmark_count": 0,
                                 "disabled_addons_ids": [],
                                 "geo_city": "Rome",
                                 "os": "Windows_NT",
                                 "subsession_length": 3785,
                                 "tab_open_count": 46,
                                 "total_uri": 791,
                                 "unique_tlds": 11,
                                 "installed_addons": ["e10srollout@mozilla.org",
                                                      "firefox@getpocket.com"],
                                 "locale": "it-IT"}
             }


def test_profile_fetcher_returns_dict(test_ctx):
    fetcher = ProfileFetcher(test_ctx)

    mock_data = MOCK_DATA['profile']
    mock_profile_controller = MockProfileController(mock_data)
    fetcher.set_client(mock_profile_controller)

    # Note that active_addons in the raw JSON source is remapped to
    # 'installed_addons'
    assert fetcher.get("random-client-id") == MOCK_DATA['expected_result']


def test_dont_crash_without_active_addons(test_ctx):
    mock_data = copy.deepcopy(MOCK_DATA['profile'])
    del mock_data['active_addons']
    mock_profile_controller = MockProfileController(mock_data)

    fetcher = ProfileFetcher(test_ctx)
    fetcher.set_client(mock_profile_controller)

    expected = copy.deepcopy(MOCK_DATA['expected_result'])
    expected['installed_addons'][:] = []
    assert fetcher.get("random-client-id") == expected


def test_crashy_profile_controller(test_ctx, monkeypatch):
    def mock_boto3_resource(*args, **kwargs):
        class ExceptionRaisingMockTable:
            def __init__(self, tbl_name):
                pass

            def get_item(self, *args, **kwargs):
                raise Exception

        class MockDDB:
            pass
        mock_ddb = MockDDB()
        mock_ddb.Table = ExceptionRaisingMockTable
        return mock_ddb

    monkeypatch.setattr(boto3, 'resource', mock_boto3_resource)

    pc = ProfileController(test_ctx, 'us-west-2', 'taar_addon_data_20180206')
    assert pc.get_client_profile("exception_raising_client_id") is None


def test_profile_controller(test_ctx, monkeypatch):
    def mock_boto3_resource(*args, **kwargs):
        some_bytes = zlib.compress(json.dumps({'key': "with_some_data"}).encode('utf8'))

        class ValueObj:
            value = some_bytes

        class MockTable:
            def __init__(self, tbl_name):
                pass

            def get_item(self, *args, **kwargs):
                value_obj = ValueObj()
                response = {'Item': {'json_payload': value_obj}}
                return response

        class MockDDB:
            pass
        mock_ddb = MockDDB()
        mock_ddb.Table = MockTable
        return mock_ddb

    monkeypatch.setattr(boto3, 'resource', mock_boto3_resource)

    pc = ProfileController(test_ctx, 'us-west-2', 'taar_addon_data_20180206')
    jdata = pc.get_client_profile("exception_raising_client_id")
    assert jdata == {'key': 'with_some_data'}



"""
A Context is a customizable namespace.

It works like a regular dictionary, but allows you to set a delegate
explicitly to do attribute lookups.

The primary benefit is that the context has a .child() method which
lets you 'lock' a dictionary and clobber the namespace without
affecting parent contexts.

In practice this makes testing easier and allows us to specialize
configuration information as we pass the context through an object
chain.
"""

# Clobber the Context name to prevent messy name collisions
from srgutil.context import default_context as _default_context


def default_context():
    ctx = _default_context()
    from taar.recommenders import CollaborativeRecommender
    from taar.recommenders import SimilarityRecommender
    from taar.recommenders import LocaleRecommender

    # Note that the EnsembleRecommender is *not* in this map as it
    # needs to ensure that the recommender_map key is installed in the
    # context
    ctx['recommender_factory_map'] = {'collaborative': lambda: CollaborativeRecommender(ctx.child()),
                                      'similarity': lambda: SimilarityRecommender(ctx.child()),
                                      'locale': lambda: LocaleRecommender(ctx.child())}

    return ctx



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from moto import mock_s3
import boto3

import json


from taar.recommenders import LocaleRecommender
from taar.recommenders.lazys3 import LazyJSONLoader
from taar.recommenders.locale_recommender import ADDON_LIST_BUCKET, ADDON_LIST_KEY


FAKE_LOCALE_DATA = {
    "te-ST": [
        "{1e6b8bce-7dc8-481c-9f19-123e41332b72}", "some-other@nice-addon.com",
        "{66d1eed2-a390-47cd-8215-016e9fa9cc55}", "{5f1594c3-0d4c-49dd-9182-4fbbb25131a7}"
    ],
    "en": [
        "some-uuid@test-addon.com", "other-addon@some-id.it"
    ]
}


def install_mock_data(ctx):
    ctx = ctx.child()
    conn = boto3.resource('s3', region_name='us-west-2')

    conn.create_bucket(Bucket=ADDON_LIST_BUCKET)
    conn.Object(ADDON_LIST_BUCKET, ADDON_LIST_KEY).put(Body=json.dumps(FAKE_LOCALE_DATA))
    ctx['locale_mock_data'] = LazyJSONLoader(ctx,
                                             ADDON_LIST_BUCKET,
                                             ADDON_LIST_KEY)

    return ctx


@mock_s3
def test_can_recommend(test_ctx):
    ctx = install_mock_data(test_ctx)
    r = LocaleRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})
    assert not r.can_recommend({"locale": []})

    # Check that we can recommend if the user has at least an addon.
    assert r.can_recommend({"locale": "en"})


@mock_s3
def test_can_recommend_no_model(test_ctx):
    ctx = install_mock_data(test_ctx)
    r = LocaleRecommender(ctx)

    # We should never be able to recommend if something went
    # wrong with the model.
    assert not r.can_recommend({})
    assert not r.can_recommend({"locale": []})
    assert not r.can_recommend({"locale": "it"})


@mock_s3
def test_recommendations(test_ctx):
    """Test that the locale recommender returns the correct
    locale dependent addons.

    The JSON output for this recommender should be a list of 2-tuples
    of (GUID, weight).
    """
    ctx = install_mock_data(test_ctx)
    r = LocaleRecommender(ctx)
    recommendations = r.recommend({"locale": "en"}, 10)

    # Make sure the structure of the recommendations is correct and that we
    # recommended the the right addon.
    assert isinstance(recommendations, list)
    assert len(recommendations) == len(FAKE_LOCALE_DATA["en"])

    # Make sure that the reported addons are the one from the fake data.
    for (addon_id, weight) in recommendations:
        assert 1 == weight
        assert addon_id in FAKE_LOCALE_DATA["en"]


@mock_s3
def test_recommender_extra_data(test_ctx):
    # Test that the recommender uses locale data from the "extra"
    # section if available.
    def validate_recommendations(data, expected_locale):
        # Make sure the structure of the recommendations is correct and that we
        # recommended the the right addon.
        assert isinstance(data, list)
        assert len(data) == len(FAKE_LOCALE_DATA[expected_locale])

        # Make sure that the reported addons are the one from the fake data.
        for (addon_id, weight) in data:
            assert addon_id in FAKE_LOCALE_DATA[expected_locale]
            assert 1 == weight

    ctx = install_mock_data(test_ctx)
    r = LocaleRecommender(ctx)
    recommendations = r.recommend({}, 10, extra_data={"locale": "en"})
    validate_recommendations(recommendations, "en")

    # Make sure that we favour client data over the extra data.
    recommendations = r.recommend({"locale": "en"}, 10, extra_data={"locale": "te-ST"})
    validate_recommendations(recommendations, "en")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from srgutil.interfaces import IMozLogging
import itertools
from .base_recommender import AbstractRecommender
from .lazys3 import LazyJSONLoader


S3_BUCKET = 'telemetry-parquet'
ENSEMBLE_WEIGHTS = 'taar/ensemble/ensemble_weight.json'


class WeightCache:
    def __init__(self, ctx):
        self._ctx = ctx

        if 'ensemble_weights' in self._ctx:
            self._weights = self._ctx['ensemble_weights']
        else:
            self._weights = LazyJSONLoader(self._ctx,
                                           S3_BUCKET,
                                           ENSEMBLE_WEIGHTS)

    def getWeights(self):
        return self._weights.get()[0]['ensemble_weights']


class EnsembleRecommender(AbstractRecommender):
    """
    The EnsembleRecommender is a collection of recommenders where the
    results from each recommendation is amplified or dampened by a
    factor.  The aggregate results are combines and used to recommend
    addons for users.
    """
    def __init__(self, ctx):
        self.RECOMMENDER_KEYS = ['collaborative', 'similarity', 'locale']
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger('taar.ensemble')

        assert 'recommender_factory' in self._ctx

        self._init_from_ctx()

    def _init_from_ctx(self):
        # Copy the map of the recommenders
        self._recommender_map = {}

        recommender_factory = self._ctx['recommender_factory']
        for rkey in self.RECOMMENDER_KEYS:
            self._recommender_map[rkey] = recommender_factory.create(rkey)

        self._weight_cache = WeightCache(self._ctx.child())

    def can_recommend(self, client_data, extra_data={}):
        """The ensemble recommender is always going to be
        available if at least one recommender is available"""
        result = sum([self._recommender_map[rkey].can_recommend(client_data)
                      for rkey in self.RECOMMENDER_KEYS])
        self.logger.info("Ensemble can_recommend: {}".format(result))
        return result

    def recommend(self, client_data, limit, extra_data={}):
        """
        Ensemble recommendations are aggregated from individual
        recommenders.  The ensemble recommender applies a weight to
        the recommendation outputs of each recommender to reorder the
        recommendations to be a better fit.

        The intuitive understanding is that the total space of
        recommended addons across all recommenders will include the
        'true' addons that should be recommended better than any
        individual recommender.  The ensemble method simply needs to
        weight each recommender appropriate so that the ordering is
        correct.
        """
        self.logger.info("Ensemble recommend invoked")
        preinstalled_addon_ids = client_data.get('installed_addons', [])

        # Compute an extended limit by adding the length of
        # the list of any preinstalled addons.
        extended_limit = limit + len(preinstalled_addon_ids)

        flattened_results = []
        ensemble_weights = self._weight_cache.getWeights()

        for rkey in self.RECOMMENDER_KEYS:
            recommender = self._recommender_map[rkey]

            if recommender.can_recommend(client_data):
                raw_results = recommender.recommend(client_data,
                                                    extended_limit,
                                                    extra_data)
                reweighted_results = []
                for guid, weight in raw_results:
                    item = (guid, weight * ensemble_weights[rkey])
                    reweighted_results.append(item)
                flattened_results.extend(reweighted_results)

        # Sort the results by the GUID
        flattened_results.sort(key=lambda item: item[0])

        # group by the guid, sum up the weights for recurring GUID
        # suggestions across all recommenders
        guid_grouper = itertools.groupby(flattened_results, lambda item: item[0])

        ensemble_suggestions = []
        for (guid, guid_group) in guid_grouper:
            weight_sum = sum([v for (g, v) in guid_group])
            item = (guid, weight_sum)
            ensemble_suggestions.append(item)

        # Sort in reverse order (greatest weight to least)
        ensemble_suggestions.sort(key=lambda x: -x[1])

        filtered_ensemble_suggestions = [(guid, weight) for (guid, weight)
                                         in ensemble_suggestions
                                         if guid not in preinstalled_addon_ids]

        results = filtered_ensemble_suggestions[:limit]

        log_data = (client_data['client_id'],
                    str(ensemble_weights),
                    str([r[0] for r in results]))
        self.logger.info("client_id: [%s], ensemble_weight: [%s], guids: [%s]" % log_data)
        return results



from setuptools import find_packages, setup

setup(
    name='mozilla-taar3',
    use_scm_version=False,
    version='0.3.4',
    setup_requires=['setuptools_scm', 'pytest-runner'],
    tests_require=['pytest'],
    include_package_data = True,
    packages=find_packages(exclude=['tests', 'tests/*']),
    description='Telemetry-Aware Addon Recommender',
    author='Mozilla Foundation',
    author_email='fx-data-dev@mozilla.org',
    url='https://github.com/mozilla/taar',
    license='MPL 2.0',
    install_requires=[
        'numpy',
        'requests',
        'thriftpy',
        'six',
        'ply',
        'boto3',
        'scipy'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment :: Mozilla',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    entry_points="""
    [taarapi_app]
    app=taar.plugin:configure_plugin
    """,
    zip_safe=False,
)
